""" Functional, deterministic augmentation on a batch of point clouds. """

from __future__ import annotations
import torch
from torch import Tensor
from torch.distributions import MultivariateNormal
from ..base.functionals import get_range, pad


def add_outliers(pointclouds: Tensor, num_points: list[int], scales: list[int]) -> Tensor:
    """ Add outliers to a batch of point clouds.

    !!! warning

        This function is a special case, as ([alto.base.functionals.sample][]). Indeed, as the
        number of outliers points and scale is unique for each point cloud, it is not possible
        to keep the batch structure throughout the entier process. Instead, the procedure is as
        follows:
        <br />
            - a list of outliers Tensors (of unique length) is created sequentially.
            <br />
            - this list is zero-padded using [alto.base.functionals.pad][].
            <br />
            - the outliers batch obtained is concatenated to the point clouds batch.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape '(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.
        num_points (list[int]): Number of outliers point for each point cloud. Of length
                                `batch_size`.
        scales (list[int]): For each point cloud, outliers points coordinates will be sampled from
                            the range `[-s * V,  s * V]`, where `s` is the `scale` value and V is
                            the radius of the smallest sphere containing the point cloud. List of
                            length `batch_size`.

    Returns:
        augmented_pointclouds: Batch of point clouds with outliers added, of shape `(batch_size,
                               N', *)`, where `*` denotes spatial coordinates and `N' = num_points
                               + max([len(x) for x in num_points])`.
    """
    centers, max_distances = get_range(pointclouds)
    outliers = list()
    for c, d, n, s in zip(centers, max_distances, num_points, scales):
        v_min, v_max = s * (c - d / 2).min(), s * (c + d / 2).max()
        sub_outliers = torch.rand(n, 3)
        sub_outliers = (v_max - v_min) * sub_outliers + v_min
        outliers.append(sub_outliers)
    outliers = pad(outliers)
    return torch.cat((pointclouds, outliers), dim=1)


def add_anisotropic_noise(pointclouds: Tensor, variances: Tensor) -> None:
    """ Sample points from a multivariate normal (gaussian) distribution parameterized by the
    given point clouds batch as mean and the given batch of variance vectors.

    !!! info

        This operation cannot be fully batched as it would requires a tremendously large covariance
        matrix. Instead, the noise is applied sequentially point after point. It is still batched
        in the sense that all points in the batch at the same index are noised simultaneously.
        That is, if `pointclouds` is a batch of shape `(B, N, 3)`, `N` operations are performed.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.
        variances (Tensor): Batch of variance vectors of shape `(batch_size, num_points, *)` where
                            `*` denotes spatial coordinates.

    Returns:
        augmented_batch: Batch of point clouds with outliers added, of shape `(batch_size,
                        N', *)`, where `*` denotes spatial coordinates. Each point has beeen
                        uniquely jitted using the provided variances.
    """
    num_points = len(pointclouds[0])
    for i in range(num_points):
        variance = variances[:, i]
        # variance shape: (batch_size, spatial_dim)
        covariance = torch.stack([torch.diag(x) for x in variance])
        # covariance shape: (batch_size, spatial_dim, spatial_dim)
        pointclouds[:, i] = MultivariateNormal(pointclouds[:, i], covariance).sample()
    return pointclouds
