from __future__ import annotations
import numpy as np
import torch
from torch import Tensor
from ..core import Transform
from . import functionals as F
from ..base.functionals import get_range


class AddRandomOutliers(Transform):

    """ Add a unique random number of outliers in each point clouds in a batch.

        !!! warning
            This Transform is a special case. Since a unique number of outliers  is added in each
            element of the batch, padding occurs under the hood. See
            [alto.base.functionals.sample][] for further details.
    """

    def __init__(self, ratio: tuple[float], scale: tuple[float], p: float = 1) -> None:
        """

        Args:
            ratio (tuple[float]): Number of outliers to add. `self.min, self.max = ratio`
                                  Then `num_outliers_min = self.min * num_points` and
                                  `num_outliers_max = self.max * num_points`

            scale (tuple[float]): Let `r` be the radius of the smallest sphere containing a
                                  point cloud (see [alto.base.functionals.get_range][]), and
                                  `self.min_scale, self.max_scale = scale`. Then for a given
                                  points cloud, outliers coordinates will be sampled uniformely
                                  in a cube centered on the point cloud center and of a random
                                  side in range `[self.min_scale * r, self.max_scale * r]`.
            p (float, optional): Probability of the transform to be applied during a call.
                                 Defaults to 1.
        """
        super().__init__(p)
        self.min_ratio, self.max_ratio = ratio
        self.min_scale, self.max_scale = scale

    def get_params(self, **data: dict[str, Tensor]) -> dict:
        """ Returns two lists: one containing the number of outliers per batch element, one
            containing the scale for coordinates sampling per batch element, from a dictionary
            whose values are batches of point clouds of shape `(batch_size, num_points,
            spatial_dim)`.

        Returns:
            dict: `dict[num_points: list[int], scales: list[int]]`. The two lists are of length
                  `batch_size`.
        """
        batch_size = Transform.get_batch_size(**data)
        num_points = Transform.get_num_points(**data)
        min_num_points, max_num_points = self.min_ratio * num_points, self.max_ratio * num_points
        num_points = np.random.randint(min_num_points, max_num_points, size=batch_size)
        scales = np.random.sample(size=batch_size)
        scales = (self.max_scale - self.min_scale) * scales + self.min_scale
        return dict(num_points=num_points, scales=scales)

    def apply(self, pointclouds: Tensor, num_points: list[int], scales: list[int]) -> Tensor:
        """ Call [alto.augment.functionals.add_outliers][] on a batch of point clouds, with
            params obtained by `self.get_params(...)`.
        """
        return F.add_outliers(pointclouds, num_points, scales)


class AddRandomAnisotropicNoise(Transform):

    """ Add anisotropic gaussian noise (move points around). Gaussian noise covariance matrix
        is unique for each point.

        !!! info
            This Transform is not fully batched as it would be too expensive. Usually in Alto, if
            a batch of shape `(B, N, 3)` is provided, one single operation is performed (!). Here,
            N operations are performed. This operation thus is a lot slower than any other in Alto.
            See [alto.augment.functionals.add_anisotropic_noise][] for further details.

        !!! warning
            As every Alto's Transform, this Transform potentially applies a random noise on several
            batches simultaneously. In that case, the uncertainy will be computed as a percentage
            of the maximum range across all point clouds in all batches.
    """

    def __init__(self, uncertainty: float, anisotropy: tuple[float], noise_uncertainty: float,
                 p: float = 1) -> None:
        """

        Args:
            uncertainty (float): Point position uncertainy, in percentage of `r`, the radius of
                                 the smallest sphere containing a point cloud (see
                                 [alto.base.functionals.get_range][]).
            anisotropy (tuple[float]): noise variance in the direction `(x, y, z)`. For instance,
                                       `(1, 1, 5)` will result in a strong anisotropic noise in the
                                       z direction.
            noise_uncertainty (float): The noise variance itself is unique in each point. This
                                       is the uncertainty on the noise itself.
            p (float, optional): Probability of the transform to be applied during a call.
                                 Defaults to 1.
        """
        super().__init__(p)
        self.uncertainty = uncertainty
        self.anisotropy  = anisotropy
        self.noise_uncertainty = noise_uncertainty

    def get_params(self, **data: dict[str, Tensor]) -> dict:
        """ Returns a tensor of unique variance vectors for each point, from a dictionary whose
            values are batches of point clouds of shape `(batch_size, num_points, spatial_dim)`.

        Returns:
            dict: `dict[variance: np.ndarray[batch_size, num_points, spatial_dim]]`.
        """
        batch_size = Transform.get_batch_size(**data)
        num_points = Transform.get_num_points(**data)
        spatial_dim = Transform.get_spatial_dim(**data)
        max_ranges = list()
        for v in data.values():
            _, max_range = get_range(v)
            max_ranges.append(max_range.max())
        absolute_max_range = max(max_ranges)
        factor = self.uncertainty * absolute_max_range
        noise = self.noise_uncertainty * factor
        scales = factor * torch.as_tensor(self.anisotropy)
        sigmas = torch.randn(batch_size, num_points, spatial_dim)
        sigmas = noise * sigmas + scales
        sigmas[sigmas < 1e-2] = 1e-2
        variances = sigmas ** 2
        return dict(variances=variances)

    def apply(self, pointclouds: Tensor, variances: Tensor) -> Tensor:
        """ Call [alto.augment.functionals.add_outliers][] on a batch of point clouds, with
            params obtained by `self.get_params(...)`.
        """
        return F.add_anisotropic_noise(pointclouds, variances)
