""" Functional, deterministic, basic operation on a batch of point clouds. """

from __future__ import annotations
from typing import Union, Sequence
import numpy as np
import torch
from torch import Tensor
from torch.nn.utils.rnn import pad_sequence
import open3d as o3d
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D


# ____________________________________________ UTILS ____________________________________________ #
# __________________ Utils functions do not have Transform class counterparts. __________________ #

def get_range(pointclouds: Tensor) -> tuple[Tensor]:
    """ Calculate the range of a batch of point clouds, that is, for each point cloud, compute the
    radius and center of the smallest sphere containing the point cloud.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.

    Returns:
        range: Center, tensor of shape `(batch_size, *)`, and radius, tensor of shape
               `(batch_size, 1)`, where `*` denotes spatial coordinates.
    """
    centers = pointclouds.mean(axis=1, keepdim=True)
    max_distances = torch.norm(pointclouds - centers, dim=2).amax(dim=1)
    return centers, max_distances


def get_fov(pointclouds: Tensor) -> Tensor:
    """ Get the range of a batch of point clouds in every spatial dimention.

    Args:
        pointcloud (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                             `*` denotes spatial coordinates.

    Returns:
        fied_of_view: Field of view, of shape `(batch_size, 2 x *)` where `*` denotes spatial
                      coordinates. *e.g.* for 2-d point clouds, field of view will contain:
                      `(batch_size, x_min, x_max, y_min, y_max)`.
    """
    spatial_dim = pointclouds.size(-1)
    fov = torch.zeros(len(pointclouds), 2 * spatial_dim)
    for i in range(spatial_dim):
        fov[:, 2 * i]     = pointclouds[:, : , i].amin(dim=1)
        fov[:, 2 * i + 1] = pointclouds[:, : , i].amax(dim=1)
    return fov


def pad(pointclouds: Sequence[Tensor], value: float=0.0) -> Tensor:
    """ Pad a list of variable length Tensors with `value`.
    This function assumes trailing dimensions and type of all the Tensors in sequences are same.

    Args:
        pointclouds (Sequence[Tensor]): Sequence of point clouds with a variable number of points.
                                        Each point cloud is of shape `(num_points, *)`, where `*`
                                        denotes spatial coordinates.
        value (float, optional): Value for padded elements.

    Returns:
        padded_batch: Padded tensor of size `(batch_size, num_points, *)`, where where `*` denotes
                      spatial coordinates, batch_size is the length of the input sequence and
                      num_points is the maximum length of input sequence elements.
    """
    return pad_sequence(pointclouds, batch_first=True, padding_value=value)


def center(pointclouds: Tensor) -> Tensor:
    """ Center a batch of point clouds on the origin.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.

    Returns:
        center: Batch of point clouds of shape `(batch_size, num_points, *)` where `*` denotes
                spatial coordinates, centered on the origin. That is, the mean over the second axis
                is `(0, 0, 0)`.
    """
    return pointclouds - pointclouds.mean(axis=1, keepdim=True)


def unite(pointclouds: Tensor) -> Tensor:
    """ Centralize and rescale a batch of point clouds so that each point cloud lies within the
    unit cube.

    Args:
        pointsclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                               `*` denotes spatial coordinates.

    Returns:
        unite_batch: Batch of point clouds of shape `(batch_size, num_points, *)` where `*` denotes
                     spatial coordinates, within the unit cube. That is, any coordinates is in the
                    range `[-0.5, 0.5]`.
    """
    max_range = (pointclouds.amax(axis=1) - pointclouds.amin(axis=1)).amax(dim=1)
    pointclouds = scale(pointclouds, 1 / max_range)
    return center(pointclouds)


# _______________________________________ BASE TRANSFORMS _______________________________________ #

def scale(pointclouds: Tensor, factors: Union[float, Sequence[float]]=1.0) -> Tensor:
    """ Multiply a batch of points coordinates by a common factor if `factors` is a float or
    by a specific value for each point cloud if `factors` is a sequence.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.
        factors (Union[float, Sequence[float]], optional):
            Factor by which to multiply each point clouds.

    Returns:
        scaled_batch: Batch of point clouds of shape `(batch_size, num_points, *)` where `*`
                      denotes spatial coordinates, where each point cloud has been scaled by the
                      given factor.
    """
    factors = torch.as_tensor(factors)
    if factors.ndim == 0:
        factors = factors.repeat(len(pointclouds))
    return pointclouds * factors.unsqueeze(dim=1)[:, :, None]


def sample(pointclouds: Tensor, indices: list[list[int]], value: float=0.0) -> Tensor:
    """ Select some points in each point clouds using the entries in indices.

    !!! warning

        Be carefull when using this function as it is a special case in this module: since a
        unique number of points can be selected for each point cloud, it is no longer possible
        to keep the batch structure throughout the entire process. Instead, the procedure is as
        follows:
        <br />
            - the batch is flatten
            <br />
            - the indices selection occurs one single time on the flatten points list (tensor)
            <br />
            - the flatten selected points are splitted according to the number of points selected in
            each point cloud
            <br />
            - the resulting sequence is padded and batched again
            (see [alto.base.functionals.pad][]).
            <br />
        This function is transparent in the sense that it takes a batch and output a batch, but be
        aware that padding occurs under the hood.
        <br />
        If you do not want this padding to occurs, you have to use this function per point cloud.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.
        indices (list[list[int]]): Indices to keep for each point cloud. Of length `batch_size`.
        value (float, optional): Value for padded elements.

    Returns:
        sampled_batch: Batch of filtered point clouds using `indices`.
    """
    flatten_points = pointclouds.flatten(start_dim=0, end_dim=1)
    num_points_per_cloud = pointclouds.size(1)
    flatten_indices = list()
    num_selected_points_per_cloud = list()
    for i, sub_indices in enumerate(indices):
        flatten_indices.extend([idx + i * num_points_per_cloud for idx in sub_indices])
        num_selected_points_per_cloud.append(len(sub_indices))
    flatten_indices = torch.as_tensor(flatten_indices)
    sampled_flatten_points = flatten_points.index_select(0, flatten_indices)
    sampled_pointclouds_list = sampled_flatten_points.split(num_selected_points_per_cloud)
    return pad(sampled_pointclouds_list, value)


# __________________________________________ VISUALIZE __________________________________________ #

def __plot_open3d(pointcloud: Tensor, point_size: float=0.01) -> None:
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(pointcloud.numpy())
    o3d.visualization.RenderOption.point_size = point_size
    mesh_frame = o3d.geometry.TriangleMesh.create_coordinate_frame()
    o3d.visualization.draw_geometries([mesh_frame, pcd])


def __plot_matplotlib(
    pointcloud: Tensor,
    ax: Axes3D=None,
    elev: float=90,
    azim: float=0,
    figsize: tuple[int]=(10, 10),
    size: float=0.1,
    colormap='cividis'
) -> None:
    """ A simple 3d point cloud plot.
    Args:
        point_cloud (np.ndarray): Array of shape (num_points, 3).
        ax (Axes3D, optional): If the plot is to be integrated within a figure, an ax can be
                               specified. If None, an axe will be created and display will be
                               handled.
        elev, azim (floats, optionals): Camera position. Angles in degrees.
        figsize (Tuple[int]): If ax is None, a figure will be created. This figures's width and
                              height can be specified here.
    """
    pointcloud = pointcloud.cpu().numpy()
    plot_later = True
    if ax is None:
        plot_later = False
        fig = plt.figure(figsize=figsize)
        ax = Axes3D(fig, elev=elev, azim=azim, auto_add_to_figure=False)
        fig.add_axes(ax)
    plt.set_cmap(colormap)
    ax.scatter(pointcloud[:, 0], pointcloud[:, 1], pointcloud[:, 2], c=pointcloud[:, 2], s=size)
    if not plot_later:
        plt.show()


def plot(pointcloud: Tensor, backend: str='matplotlib', **kwargs: dict) -> None:
    """ A simple plot utility.

    !!! tip
        The open3d backend uses the GPU to render the point cloud. It should be your default
        choice. Be aware though that it could cause some issues when used inside a notebook.

    Args:
        pointcloud (Tensor): A single point cloud, of shape `(num_points, *)` where `*` denotes
                             spatial coordinates.
        backend (str, optional): One of `('open3d', 'matplotlib')`. `'open3d'` backend is
                                 interactive and GPU backed.
        kwargs (dict, optional): Optional keywords arguments, depend on the backend. For
                                 `'open3d'`, the only available argument is `point_size`.
                                 For `'matplotlib'`, available arguments are `ax` (in cas you wanna
                                 handle subplots), `elev` and 'azim` (camera position), `figsize`,
                                 `size`, and  `colormap`.

    """
    plot_fn = __plot_open3d if backend == 'open3d' else __plot_matplotlib
    plot_fn(pointcloud, **kwargs)


def grid_plot(
    pointclouds: Tensor,
    elev: int=90,
    azim: int=0,
    figsize: tuple[int]=None,
    axis: bool=False,
    size: float=0.1,
    colormap='cividis'
) -> None:
    """ Plot a batch of points cloud in grid.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.
        elev (int, optional): Camera elevation.
        azim (int, optional): Camera azimuth.
        figsize (tuple[int], optional): Total figure size. If None, will be set to `(8 * ncols,
                                        8 * nrows)`, where ncols is 2 if `batch_size` < 4, 4
                                        otherwise, and `nrows` is `ceil(nplots / ncols)`.
        axis (bool, optional): Wether to show awis or not.
        size (float, optional): Point size.
        colormap (str, optional):
    """
    nplots = len(pointclouds)
    ncols = 2 if nplots < 4 else 4
    nrows = int(np.ceil(nplots / ncols))
    if figsize is None:
        figsize = (8 * ncols, 8 * nrows)
    fig = plt.figure(figsize=figsize)
    for i in range(nplots):
        ax = fig.add_subplot(nrows, ncols, i + 1, projection='3d', elev=elev, azim=azim)
        plot(pointclouds[i], ax=ax, size=size, colormap=colormap)
        if not axis:
            ax.set_axis_off()
    plt.tight_layout()
    plt.show()
