from __future__ import annotations
from typing import Sequence
import numpy as np
from torch import Tensor
from ..core import Transform
from . import functionals as F


class RandomScale(Transform):

    """ Randomly scale a batch of point clouds, with a unique scale factor for each point cloud.
    """

    def __init__(self, range: tuple[float], p: float = 1) -> None:
        """

        Args:
            range (tuple[float]): min scale value, max scale value.
            p (float, optional): Probability of the transform to be applied during a call.
                                 Defaults to 1.
        """
        super().__init__(p)
        self.min, self.max = range

    def get_params(self, **data: dict[str, Tensor]) -> dict:
        """ Return a list of scale values sampled uniformely in [self.min, self.max] (ie range).

        Returns:
            params: `dict['factors': np.ndarray[batch_size]]`.
        """
        factors = np.random.random_sample(size=Transform.get_batch_size(**data))
        factors = (self.max - self.min) * factors + self.min
        return dict(factors=factors.astype(np.float32))

    def apply(self, pointclouds: Tensor, factors: Sequence[float]) -> Tensor:
        """ Call [alto.base.functionals.scale][] on a batch of point clouds, with params
            obtained by `self.get_params(...)`.
        """
        return F.scale(pointclouds, factors)


class RandomSample(Transform):

    """ Randomly sample points from a batch of point clouds. Select a unique number of points
        in each point clouds.

        !!! warning
            This Transform is a special case. Since a unique number of point is selected in each
            element of the batch, padding occurs under the hood. See
            [alto.base.functionals.sample][] for further details.
    """

    def __init__(self, range: tuple[float], value: float=0.0, p: float = 1) -> None:
        """

        Args:
            range (tuple[float]): Number of points to keep, in percentage of the current number
                                  of points. *e.g.*: if the provided batch consists of clouds of
                                  100 points and range=(0.5, 1.0), the resulting batch will have
                                  between 50 and 100 points.
            value (float, optional): Padding value. Defaults to 0.0.
            p (float, optional): Probability of the transform to be applied during a call.
                                 Defaults to 1.
        """
        super().__init__(p)
        self.min, self.max = range
        self.value = value

    def get_params(self, **data: dict[str, Tensor]) -> dict:
        """ Return a list of list of indices to be kept from a dictionary whose values are batches
            of point clouds of shape `(batch_size, num_points, spatial_dim)`.

        Returns:
            params: `dict[indices: list[list[int]]`. The wrapping list is of length `batch_size`
                    and every element inside is a list of integers in range [0, num_points] and
                    of length in range `[self.min * num_points, self.max * num_points]`.
        """
        batch_size = Transform.get_batch_size(**data)
        num_points = Transform.get_num_points(**data)
        min_num_points, max_num_points = self.min * num_points, self.max * num_points
        indices = list()
        for _ in range(batch_size):
            size = np.random.randint(min_num_points, max_num_points)
            indices.append(np.random.choice(list(range(num_points)), size=size))
        return dict(indices=indices, value=self.value)

    def apply(self, pointclouds: Tensor, indices: list[list[int]], value: float=0.0) -> Tensor:
        """ Call [alto.base.functionals.sample][] on a batch of point clouds, with params
            obtained by `self.get_params(...)`.
        """
        return F.sample(pointclouds, indices, value)
