""" Functional, deterministic translation and rotation on a batch of point clouds.
    Rotation functions are adapted from pytorch3d.
"""

from __future__ import annotations
from typing import Union, Sequence
import torch
from torch import Tensor


def translate(
    pointclouds: Tensor,
    values: Union[Sequence[float], Sequence[Sequence[float]]]=(0.0, 0.0, 0.0)
) -> Tensor:
    """ Translate a batch of points coordinates by a common value if `values` is a triplet of
    floats or by a specific value for each point cloud if `values` is a sequence of triplets.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.
        values (Union[Sequence[float], Sequence[Sequence[float]]], optional):
            Value by which to translate each point clouds.

    Returns:
        translated_batch: Batch of point clouds of shape `(batch_size, num_points, *)` where `*`
                          denotes spatial coordinates, where each point cloud has been translated
                          by the given value.
    """
    values = torch.as_tensor(values)
    if values.ndim == 1:
        values = values.repeat(len(pointclouds), 1)
    return pointclouds + values[:, None, :]


def __degrees_to_radians(degrees: Tensor) -> Tensor:
    """ A basic batched degree to radians conversion.

    Args:
        degrees (Tensor[int]): Batch of angles in degrees.

    Returns:
        Tensor[float]: Batch of angles in radians.
    """
    return torch.pi / 180 * degrees


def __axis_angle_rotation(axis: str, angle: Tensor) -> Tensor:
    """ Return the rotation matrices for one of the rotations about an axis of which Euler angles
    describe, for each value of the angle given.

    Args:
        axis (str): Axis label: 'x', 'y', or 'z'.
        angle (Tensor): Euler angles in radians.

    Returns:
        Tensor: Rotation matrices as tensor of shape (..., 3, 3).
    """
    cos = torch.cos(angle)
    sin = torch.sin(angle)
    one = torch.ones_like(angle)
    zero = torch.zeros_like(angle)
    if axis == 'x':
        R_flat = (one, zero, zero, zero, cos, -sin, zero, sin, cos)
    elif axis == 'y':
        R_flat = (cos, zero, sin, zero, one, zero, -sin, zero, cos)
    elif axis == 'z':
        R_flat = (cos, -sin, zero, sin, cos, zero, zero, zero, one)
    return torch.stack(R_flat, -1).reshape(angle.shape + (3, 3))


def __euler_angles_to_matrix(euler_angles: Tensor, convention: str) -> Tensor:
    """ Convert rotations given as Euler angles in radians to rotation matrices.

    Args:
        euler_angles: Euler angles in radians as tensor of shape (..., 3).
        convention: Convention string of three lowercase letters from {'x', 'y', and 'z'}.

    Returns:
        Tensor: Rotation matrices as tensor of shape (..., 3, 3).
    """
    matrices = list()
    for c, e in zip(convention, torch.unbind(euler_angles, -1)):
        matrices.append(__axis_angle_rotation(c, e))
    return torch.matmul(torch.matmul(matrices[0], matrices[1]), matrices[2])


def rotate(
    pointclouds: Tensor,
    euler_angles: Union[Sequence[int], Sequence[Sequence[int]]],
    convention: str
) -> Tensor:
    """ Rotate a batch of Tensors using Euler angles with the specified convention.
    If `euler_angles` is of size `3`, the same rotation is applied to each point cloud.
    If `euler_angles` is a list of triplets of same length as `pointclouds`, each point cloud
    is uniquely rotated.

    Args:
        pointclouds (Tensor): Batch of point clouds of shape `(batch_size, num_points, *)` where
                              `*` denotes spatial coordinates.
        euler_angles (Union[Sequence[int], Sequence[Sequence[int]]]):
            Either a triplet of Euler angles or a sequence of triplets of length equals to the
            batch one.
        convention (str): Convention string of three lowercase letters from `{'x', 'y', 'z'}`.

    Returns:
        rotated_batch: Rotated batch of point clouds, of shape `(batch_size, num_points, *)` where
                       `*` denotes spatial coordinates.
    """
    euler_angles = torch.as_tensor(euler_angles)
    if euler_angles.ndim == 1:
        rotation_matrix = __euler_angles_to_matrix(euler_angles, convention)
    else:
        rotation_matrix = [__euler_angles_to_matrix(a, convention) for a in euler_angles]
        rotation_matrix = torch.stack(rotation_matrix)
    return torch.matmul(pointclouds, rotation_matrix)
