from __future__ import annotations
from typing import Sequence
import numpy as np
from torch import Tensor
from ..core import Transform
from . import functionals as F


class RandomTranslate(Transform):

    """ Translate each point cloud in a batch by a random unique value. """

    def __init__(self, range: tuple[float], p: float = 1) -> None:
        """

        Args:
            range (tuple[float]): min scale value, max scale value.
            p (float, optional): Probability of the transform to be applied during a call.
                                 Defaults to 1.
        """
        super().__init__(p)
        self.min, self.max = range

    def get_params(self, **data: dict[str, Tensor]) -> dict:
        """ Return a list of translation vectors to be kept from a dictionary whose values are
        batches of point clouds of shape `(batch_size, num_points, spatial_dim)`.

        Returns:
            dict: `dict[values: np.ndarray[batch_size, spatial_dim]]`.
        """
        batch_size = Transform.get_batch_size(**data)
        spatial_dim = Transform.get_spatial_dim(**data)
        values = np.random.random_sample(size=(batch_size, spatial_dim))
        values = (self.max - self.min) * values + self.min
        return dict(values=values.astype(np.float32))

    def apply(self, pointclouds: Tensor, values: Sequence[float]) -> Tensor:
        """ Call [alto.so3.functionals.translate][] on a batch of point clouds, with params
            obtained by `self.get_params(...)`.
        """
        return F.translate(pointclouds, values)


class RandomRotate(Transform):

    """ Translate each point cloud in a batch by a random unique rotation.
        Rotation matrix are applied from random Euler's angles, in the provided range, according
        to the provided convention.
    """

    def __init__(self, range: tuple[float], convention: str, p: float = 1) -> None:
        """

        Args:
            range (tuple[float]): `(min1, max1, min2, max2, min3, max3)`. Meanings of `1`, `2`,
                                  and `3` are determined by `convention`.
            convention (str): Convention string of three lowercase letters from `{'x', 'y',
                              and 'z'}`.
            p (float, optional): Probability of the transform to be applied during a call.
                                 Defaults to 1.
        """
        super().__init__(p)
        self.range = range
        self.convention = convention

    def get_params(self, **data: dict[str, Tensor]) -> dict:
        """ Returns a list of Euler angles triplets from a dictionary whose values are batches
            of point clouds of shape `(batch_size, num_points, spatial_dim)`.

        Returns:
            dict: `dict[euler_angles: list[list[float]]]`. The wrapping list is of length
                  `batch_size` all its elements are triplets of floats (Euler angles in the
                  convention `self.convention`).
        """
        batch_size = Transform.get_batch_size(**data)
        euler_angles = list()
        for i in range(0, len(self.range), 2):
            angle_min, angle_max = self.range[i], self.range[i + 1]
            if angle_min == angle_max:
                euler_angles.append(batch_size * (angle_min, ))
            else:
                euler_angles.append(np.random.randint(angle_min, angle_max, size=batch_size))
        euler_angles = np.stack(euler_angles).transpose()
        return dict(euler_angles=euler_angles)

    def apply(self, pointclouds: Tensor, euler_angles: Sequence[Sequence[int]]) -> Tensor:
        """ Call [alto.so3.functionals.rotate][] on a batch of point clouds, with params
            obtained by `self.get_params(...)`.
        """
        return F.rotate(pointclouds, euler_angles, self.convention)
